#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright © 2004-2005 Guido Amoruso

This file is part of xml2ly

xml2ly is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

xml2ly is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""


from musicxml import *
from reader import *

import string
import sys


outfile = sys.stdout
errfile = sys.stderr

debug_on = False


def set_output_file (out):
	global outfile
	outfile = out


def write (string):
	outfile.write (string)


def nl (indent = 0):
	global indent_level, indenter
	
	assert indent_level >= 0
	indent_level += indent
	s = ''
	for i in range (indent_level):
		s += indenter
		
	outfile.write ('\n' + s)


def sp ():
	outfile.write (' ')


def error (string):
	errfile.write (string)


def debug (string):
	if debug_on:
		write (string)


def rewind (string):
	for i in range (len (string)):
		outfile.seek (-1, 1)


# options
print_slurs = True
print_lyrics = False
lyrics_by_note = True
relative_pitch = True
relative_duration = True

sharpeye_bug = False

lily_version = '2.7.10'
bar_check = 'start'
bar_comment_modulo = 5
slurs_next_to_notes = True
indenter = '  '
indent_level = 0


def dump_score (score):
	global sharpeye_bug

	try:
		if 'SharpEye' in score.identification.encoding['software'][0]:
			sharpeye_bug = True
	except:
		pass

	write ('\\version "' + lily_version + '"')
	nl ()
	nl ()

	write ('\\header {')
	nl (1)
	nl ()
	try:
		score.identification.dump ()
	except:
		pass

	try:	
		score.work.dump ()
	except:
		pass

	nl (-1)
	write ('}')
	nl ()
	nl ()

	for i in score.parts:
		i.dump ()

	nl ()
	write ('\\score {')
	nl (1)
	write ('<<')
	nl (1)
	for i in score.parts:
		write ('\\' + string2ly (i.id))
		nl ()

	rewind (indenter)
	write ('>>')
	nl (-2)

	write ('}')
	nl ()

Score.dump = dump_score


def dump_identification (identification):
	write ('composer = "')
	try:
		write (identification.creator + '"')
	except AttributeError:
		write ('"')
	nl ()
	
Identification.dump = dump_identification


def dump_work (work):
	write ('title = "')
	try:
		write (work.work_title + '"')
	except AttributeError:
		write ('"')
	nl ()

	write ('opus = "')
	try:
		write (work.work_number + '"')
	except AttributeError:
		write ('"')
	nl ()
	
Work.dump = dump_work



def dump_part (part):
	voices_on_staff = {1 : [], 2 : []}
	
	write (string2ly (str (part.id)) + ' = { ')
	nl (1)

	voices = part.voices.keys ()
	voices.sort ()

	for v in voices:
		if part.voices[v].staff () == 1:
			voices_on_staff[1].append (v)
		elif part.voices[v].staff () == 2:
			voices_on_staff[2].append (v)


	if part.staves == 2:
		write ('\context PianoStaff <<')	
	elif part.staves == 1:
		write ('\\new Staff <<')

	nl (1)

	staffs = {1 : 'up', 2 : 'down'}
	for i in [1, 2]:
		if part.staves == 2:
			write ('\context Staff = ' + staffs[i] + ' <<')
			nl (1)
			write ('\clef ' + clef2ly (part.clef[i]))
			nl ()
			write ('\key ' + key2ly (part.key))
			nl ()
			write ('\\time ' + time2ly (part.time))
			nl ()

		for v in voices_on_staff[i]:
			part.voices[v].dump ()

			if len (voices_on_staff[i]) > 1 and v != max (voices_on_staff[i]):
				nl ()
				write ('\\\\')
				nl ()
				nl ()

		if part.staves == 2:
			rewind (indenter)
			write ('>>')
			nl (-1)
			nl ()

	rewind (indenter)
	
	write ('>>')
	nl (-1)
	rewind (indenter)

	write ('}')
	nl (-1)
	nl ()

Part.dump = dump_part


def dump_voice (voice):
	measures = voice.measures.keys ()
	measures.sort ()

	if relative_pitch:
		write ('\\relative ')
		first_chord = voice.measures[measures[0]].first_chord ()
		if first_chord and clef2ly (first_chord.notes[0].clef) == 'bass':
			write ('c ')
		else:
			write ("c' ")

	write ('{')
	nl (1)

	# KLUDGE: pickup measure....
	for i in range (len (voice.part) + 1):
		if voice.has_measure (i):
			voice.measures[i].dump ()
		else:
			if i > 0:
				if voice.part.is_pickup () and sharpeye_bug:
					if ((i - 1) % bar_comment_modulo) == 0:
						write ('% bar ' + str (i - 1))
						nl ()
				else:
					if (i % bar_comment_modulo) == 0:
						write ('% bar ' + str (i))
						nl ()
				if bar_check == 'start':
					write ('| ')
				write ('s1*' + time2ly (voice.part.time_at_measure (i)))
				nl ()
				if bar_check == 'end':
					write (' |')
			elif voice.part.is_pickup ():
				try:
					duration = voice.part.is_pickup ()
					divisions = voice.part.divisions_at_measure (i)[0].number
					timeslices = [{'duration' : duration, 'divisions' : divisions}]
					write ('\\partial ' + partial2ly (timeslices) + ' ')
					write (silence2ly (timeslices))
					nl ()
				except TypeError:
					pass

	rewind (indenter)
	write ('}')
	nl (-1)

Voice.dump = dump_voice


def dump_measure (measure):
	global indent_level
	pad = 0
	measure_length = len (measure)
	last_chord = measure.last_chord ()

	if measure.voice.part.is_pickup () and sharpeye_bug:
		if ((measure.number - 1) % bar_comment_modulo) == 0:
			write ('% bar ' + str (measure.number - 1))
			nl ()
	else:
		if (measure.number % bar_comment_modulo) == 0:
			write ('% bar ' + str (measure.number))
			nl ()

	if bar_check == 'start':
		indent_level += 1
		
		if measure.number > 0:
			write ('| ')
		else:
			write ('  ')

	if measure.first_chord () and (measure.first_chord ().moment > 0):
		debug ('SKIP: <<< ')
		write (silence2ly (measure.timeslices (0, measure.first_chord ().moment)))
		debug (' >>> ')

	for i in measure.events:
		# pickup measures (SharpEye bug: pickup measures have number == 1)
		if measure.number == 0 or (sharpeye_bug and measure.number == 1):
			if not isinstance (i, Chord) or not i.is_first_chord_in_measure ():
				i.dump ()
			else:
				pad = 0
				if last_chord:
					pad = measure_length - (last_chord.moment + last_chord.duration)
					if (pad):
						debug ('PARTIAL <<< ')
						timeslices = measure.timeslices (0, last_chord.moment + last_chord.duration)
						write ('\\partial ' + partial2ly (timeslices))
						nl ()
						debug (' >>> ')
				i.dump ()
		else:
			i.dump ()

		
		if isinstance (i, Chord) and isinstance (i.notes[0], Note) and i.next_chord ():
			pad = 0
			if i.measure == i.next_chord ().measure:
				next_start = i.next_chord_not_grace_not_cue ().moment
				pad = (i.moment + i.duration) - next_start
				if pad:
					debug ('SKIPMID: <<< ')
					timeslices = measure.timeslices (i.moment + i.duration, next_start)
					write (silence2ly (timeslices))
					debug (' >>> ')

	pad = 0
	if last_chord:
		pad = measure_length - (last_chord.moment + last_chord.duration)
		if (pad):
			if measure.number > 1 or (measure.number == 1 and not sharpeye_bug):
				debug ('SKIPEND: <<< ')
				timeslices = measure.timeslices (last_chord.moment + last_chord.duration,
								 measure_length)
				write (silence2ly (timeslices))
				debug (' >>> ')
	else:
		debug ('SKIPTOTAL: <<< ')
		timeslices = measure.timeslices (0, measure_length)
		write (silence2ly (timeslices))
		debug (' >>> ')
		

	if bar_check == 'end':
		write (' |')

	if bar_check == 'start':
		indent_level -= 1

	nl ()

Measure.dump = dump_measure


def dump_event (event):
	pass

Event.dump = dump_event


def dump_chord (chord):
	duration = ''
	prev_chord = chord.prev_chord ()
	next_chord = chord.next_chord ()

	if len (chord.measure.voice.part.voices) > 1:
		staffs = {1 : 'up', 2: 'down'}
		stems = {'up' : 'Up', 'down' : 'Down', None : 'Neutral'}
		
		if not prev_chord:
			if chord.staff != chord.measure.voice.staff ():
				write ('\change Staff = ' + staffs[chord.staff])
				sp ()
			write ('\\stem' + stems[chord.notes[0].stem])
			sp ()
		else:
			prev_clef = chord.prev_clef ()
			if prev_chord.staff != chord.staff or (prev_clef and prev_clef.staff != chord.staff):
				write ('\change Staff = ' + staffs[chord.staff])
				sp ()
			if prev_chord.notes[0].stem != chord.notes[0].stem:
				write ('\\stem' + stems[chord.notes[0].stem])
				sp ()

	if chord.is_grace ():
		if not prev_chord or not prev_chord.is_grace ():
			write ('\grace { ')
	
	if hasattr (chord.notes[0], 'time_modification'):
		prev = chord.prev_chord_not_grace_not_cue ()
		if prev is None \
		   or not hasattr (prev.notes[0], 'time_modification'):
			write ('\set Voice.tupletSpannerDuration = #(ly:make-moment ')
			write (str (chord.notes[0].time_modification.normal_notes))
			sp ()
			write (type2ly (chord.notes[0].type))
			write (')')
			nl ()
			write ('\\times '                                                         \
			      + str (chord.notes[0].time_modification.normal_notes) + '/'         \
			      + str (chord.notes[0].time_modification.actual_notes) + ' {' + ' ')
			nl (1)
	
	if len (chord.notes) > 1:
		write ('<')

	for i in range (len (chord.notes)):
		step  = ''
		octave = ''
		alter = ''
		rest = ''

		if hasattr (chord.notes[i], 'pitch'):
			step = step2ly (chord.notes[i].pitch.step)
			alter = alter2ly (chord.notes[i].pitch.alter)

			if relative_pitch:
				# check whether a not-first-note-in-chord is being dumped
				if i > 0:
					interval = chord.notes[i].pitch - chord.notes[i - 1].pitch
				elif prev_chord and chord.prev_pitch ():
					interval = chord.notes[i].pitch - chord.prev_pitch ()
				else:
					pitch = Pitch ()
					pitch.step = 'C'
					pitch.alter = 0
					if clef2ly (chord.notes[i].clef) == 'bass':
						pitch.octave = 3
					else:
						pitch.octave = 4
					interval = chord.notes[i].pitch - pitch
					
				while not (-4 <= abs (interval) <= 4):
					if interval > 0:
						octave += "'"
					else :
						octave += ','
					interval += -7 * (interval / abs (interval))
			else:
				octave = octave2ly (chord.notes[i].pitch.octave)
	
		if hasattr (chord.notes[i], 'rest'):
			rest = 'r'
	
		write (step + alter + rest + octave)
		if len (chord.notes) > (i + 1):
			sp ()

	if len (chord.notes) > 1:
		write ('>')

	# BUG: duration should be output even when there is a skip at the end of the
	# previous measure, or when there is not previous measure
	if hasattr (chord.notes[0], 'type'):
		if not prev_chord                                                        \
		   or chord.notes[0].dot > 0                                             \
		   or prev_chord.notes[0].dot > 0                                        \
		   or type2ly (prev_chord.notes[0].type, prev_chord.notes[0].time)       \
		      != type2ly (chord.notes[0].type, chord.notes[0].time)              \
		   or (prev_chord.moment + prev_chord.duration < chord.moment)           \
		   or (prev_chord.measure.number != chord.measure.number):

			duration = type2ly (chord.notes[0].type, chord.notes[0].time)
	else:
		try:
			duration = '1*' + str (chord.notes[0].time.beats) + '/' + str (chord.notes[0].time.beat_type)
		except AttributeError:
			duration = '1*4/4'

	for i in range (chord.notes[0].dot):
		duration += '.'

	write (duration)
	
	# All notations are precessed here together, for all the notes in the chord.
	# Perhaps they should be spread more carefully over each note in the chord.
	articulations = []
	for n in chord.notes:
		articulations += n.articulations
	for a in articulations:
		a.dump ()

	ornaments = []
	for n in chord.notes:
		ornaments += n.ornaments
	for a in ornaments:
		a.dump ()

	technicals = []
	for n in chord.notes:
		technicals += n.technicals
	for a in technicals:
		a.dump ()

	slurs = []
	for n in chord.notes:
		slurs += n.slurs
	for a in slurs:
		a.dump ()

	# BUG: ties should be spread over notes
	ties = []
	for n in chord.notes:
		ties += n.ties
	for a in ties:
		if a.type == 'start':
			write ('~')
			break
		
		continue

	sp ()

	# direction happening at this moment
	if not chord.is_grace ():
		directions = chord.measure.directions_at (chord.moment)
		for d in directions:
			if isinstance (d, Dynamics):
				dump_dynamics (d)
			elif isinstance (d, Wedge):
				dump_wedge (d)

	if hasattr (chord.notes[0], 'time_modification'):
		next = chord.next_chord_not_grace_not_cue ()
		if next is None \
		   or not hasattr (next.notes[0], 'time_modification'):
			nl (-1)
			write ('} ')

	if chord.is_grace ():
		if not chord.next_chord ().is_grace ():
			write ('} ')

Chord.dump = dump_chord


def dump_clef (clef):	
	prev_chord = clef.prev_chord ()
	
	if len (clef.measure.voice.part.voices) > 1:
		staffs = {1: 'up', 2: 'down'}
		
		if not prev_chord:
			if clef.staff != clef.measure.voice.staff ():
				write ('\change Staff = ' + staffs[clef.staff])
				sp ()
		elif prev_chord.staff != clef.staff:
				write ('\change Staff = ' + staffs[clef.staff])
				sp ()

	write ('\clef ' + clef2ly (clef))
	nl ()

Clef.dump = dump_clef


def dump_time (time):
	write ('\\time ' + time2ly (time))
	nl ()

Time.dump = dump_time


def dump_key (key):
	prev_chord = key.prev_chord ()
	
	if len (key.measure.voice.part.voices) > 1 and key.staff != 0:
		staffs = {1 : 'up', 2: 'down'}
		
		if not prev_chord:
			if key.staff != key.measure.voice.staff ():
				write ('\change Staff = ' + staffs[key.staff])
				sp ()
		elif prev_chord.staff != key.staff:
				write ('\change Staff = ' + staffs[key.staff])
				sp ()
	
	write ('\key ' + key2ly (key))
	nl ()

Key.dump = dump_key


def dump_articulations (articulations):
	s = ''
	
	for i in articulations.values:
		for n in range (articulations.values[i]):
			s += articulation2ly (i)

	write (s)

Articulations.dump = dump_articulations


def dump_ornaments (ornaments):
	s = ''
	
	for i in ornaments.values:
		for n in range (ornaments.values[i]):
			s += ornament2ly (i)

	write (s)

Ornaments.dump = dump_ornaments


def dump_technical (technical):
	s = ''
	
	for i in technical.values:
		for n in range (technical.values[i]):
			s += technical2ly (i)

	write (s)

Technical.dump = dump_technical


def dump_slur (slur):
	s = ''

	# BUG: deeply nested (more than 2) slurs are skipped
	if slur.number > 2:
		return
	
	if slur.number == 2:
		s = '\\'

	if slur.type == 'start':
		s += '('
	elif slur.type == 'stop':
		s += ')'
	else:
		raise Exception ('UnknownSlurTypeError')

	write (s)
	

Slur.dump = dump_slur


def dump_dynamics (dynamics):
	for i in dynamics.values:
		write ('\\' + str (i) + ' ')

# Dynamics.dump = dump_dynamics


def dump_wedge (wedge):
	if wedge.type == 'crescendo':
		write ('\\< ')
	elif wedge.type == 'diminuendo':
		write ('\\> ')
	elif wedge.type == 'stop':
		write ('\\! ')

# Wedge.dump = dump_wedge


# processing utilites

def step2ly (step):
	return string.lower (step)
	

def alter2ly (alter):
	if alter == 1: alter_sign = 'is'
	elif alter == 2: alter_sign = 'isis'
	elif alter == -1: alter_sign = 'es'
	elif alter == -2: alter_sign = 'eses'
	else: alter_sign = ''

	return alter_sign


def octave2ly (octave):
	octave_sign = [",,,",     \
		       ",,",      \
		       ",",       \
		       "",        \
		       "'",       \
		       "''",      \
		       "'''",     \
		       "''''",    \
		       "'''''",   \
		       "''''''"]

	return octave_sign[octave]


def type2ly (type, time = None):
	types = {'breve'   : '1*1/2', \
		 'half'    : '2',     \
		 'quarter' : '4',     \
		 'eighth'  : '8',     \
		 '16th'    : '16',    \
		 '32nd'    : '32',    \
		 '64th'    : '64',    \
		 '128th'   : '128',   \
		 'fake'    : 'fake'   \
		}

	if time is None:
		time = Time ({'beats' : [4], 'beat-type' : [4]})

	if type == 'whole':
		return '1*' + str (time.beats[0]) + '/' + str (time.beat_type[0])
	elif type in types:
		return types[type]
	else:
		return '1*' + str (time.beats[0]) + '/' + str (time.beat_type[0])


def clef2ly (clef):
	name = ''

	if clef is None:
		return 'violin'
	
	if clef.sign == 'G':
		if clef.line == 1:
			name = 'french_violin'
		elif clef.line == 2 or clef.line == 0:
			name = 'violin'
		else:
			raise Exception ('UnknownClefWarning')

	elif clef.sign == 'F':
		if clef.line == 4 or clef.line == 0:
			name = 'bass'
		elif clef.line == 3:
			name = 'bariton'
		else:
			raise Exception ('UnknownClefWarning')

	elif clef.sign == 'C':
		if clef.line == 1:
			name = 'soprano'
		elif clef.line == 2:
			name = 'mezzosoprano'
		elif clef.line == 3 or clef.line == 0:
			name = 'alto'
		elif clef.line == 4:
			name = 'tenor'
		else:
			raise Exception ('UnknownClefWarning')

	return name


def key2ly (key):
	name = ''

	fifths = ['c', \
		  'g', 'd', 'a', 'e', 'b', 'fis', 'cis', \
		  'ces', 'ges', 'des', 'aes', 'ees', 'bes', 'f']

	if key.mode == 'minor':
		idx = key.fifths + 3
		name = fifths[idx]
		name += ' \minor'
	else:
		name = fifths[key.fifths]
		name += ' \major'

	return name


def time2ly (time):
	if time is None:
		return '4/4'
	
	return str (time.beats[0]) + '/' + str (time.beat_type[0])
	

def string2ly (s):
	subst = [[' ', '_'],       \
		 ['0', 'Zero'],    \
		 ['1', 'One'],     \
		 ['2', 'Two'],     \
		 ['3', 'Three'],   \
		 ['4', 'Four'],    \
		 ['5', 'Five'],    \
		 ['6', 'Six'],     \
		 ['7', 'Seven'],   \
		 ['8', 'Eight'],   \
		 ['9', 'Nine']]

	for i in subst:
		s = string.replace (s, i[0], i[1])

	return s


def silence2ly (timeslices):
	s = ''
	
	for i in timeslices:
		j = 0
		d = 4
		n = int (i['duration'] / i['divisions'])
		for k in range (n):
			s += 's' + str (d) + ' '
			
		x = i['duration'] % i['divisions']
		while x != 0 and j < 3:
			i['divisions'] /= 2
			d *= 2
			n = int (x / i['divisions'])
			for k in range (n):
				s += 's' + str (d) + ' '
			x = i['duration'] % i['divisions']

		return s


def partial2ly (timeslices):
	s = ''
	duration = timeslices[0]['duration']
	divisions = timeslices[0]['divisions']

	d = 4
	while (duration % divisions) != 0:
		duration *= 2
		d *= 2

	n = duration / divisions

	return str (d) + '*' + str (n)
	

def articulation2ly (articulation):
	articulations = {'accent'          : '>',    \
			 'strong-accent'   : '^',    \
			 'staccato'        : '.',    \
			 'tenuto'          : '-',    \
			 'detached-legato' : '.--',  \
			 'staccatissimo'   : '|'     \
			}

	if articulation in articulations:
		return '-' + articulations[articulation]
	# temporary trick, for an improbable articulation mark
	else:
		return '-' + '.->-^--'

def ornament2ly (ornament):
	ornaments = {'trill-mark'            : 'trill',         \
		     'turn'                  : 'turn',          \
		     'mordent'               : 'mordent',       \
		     'long-mordent'          : 'prallmordent',  \
		     'inverted-mordent'      : 'prall',         \
		     'long-inverted-mordent' : 'prallprall',    \
		     'delayed-turn'          : 'reverseturn'}
	
	if ornament in ornaments:
		return '-\\' + ornaments[ornament]
	# temporary trick
	else:
		return '-\\' + 'trill'

def tecnical2ly (technical):
	pass



def main ():
	parser = make_parser ()
	handler = MusicXMLContentHandler ()
	parser.setContentHandler (handler)
	parser.setFeature (feature_external_ges, False)

	parser.parse (open (sys.argv[1]))
	score = handler.score
	score.dump ()


if __name__ == '__main__':
	main ()

