#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright © 2004-2005 Guido Amoruso

This file is part of xml2ly

xml2ly is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

xml2ly is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""

import copy
import math
import sys

True = 1
False = 0


def set_attributes (object, attrs, treefrag):
	for a in attrs:
		key = a[0].replace ('-', '_')
		try:
			if len (a) >= 2 and callable (a[1]):
				f = a[1]
				object.__dict__[key] = f (treefrag[a[0]])
			else:
				try:
					object.__dict__[key] = treefrag[a[0]][0]
				except TypeError:
					object.__dict__[key] = treefrag[a[0]]
		except KeyError:
			if len (a) >= 2 and not callable (a[-1]):
				object.__dict__[key] = a[-1]



class Object (object):
	def __init__ (self):
		object.__init__ (self)

	def __str__ (self):
		return str (self.__dict__)

	def dump (self):
		print str (self)

	dump = dump


class Score (Object):
	def __init__ (self):
		self.parts = []
		self.identification = None
		self.work = None

	def add_part (self, part):
		part.score = self
		self.parts.append (part)


class Identification (Object):
	def __init__ (self, treefrag):
		attrs = [['creator'],    \
			 ['rights'],      \
			 ['encoding'],        \
			 ['supports'],\
			 ['source'],\
			 ['miscellaneous']]

		set_attributes (self, attrs, treefrag)


class Work (Object):
	def __init__ (self, treefrag):
		attrs = [['work-number'],    \
			 ['work-title']]

		set_attributes (self, attrs, treefrag)


class Part (Object):
	def __init__ (self, id):
		self.id = id
		self.score = None
		self.voices = {}
		self.staves = 1

		self.time = None
		self.key = None
		self.clef = {1: None, 2: None}

	def add_voice (self, voice):
		voice.part = self
		self.voices[voice.number] = voice

	def add_event (self, event):
		if isinstance (event, Time) and not self.time:
			self.time = event
		elif isinstance (event, Clef) and not self.clef[event.staff]:
			# self.clef[event.staff] = event
			self.clef[1] = event
			self.clef[2] = event
		elif isinstance (event, Key) and not self.key:
			self.key = event
		
		if event.voice not in self.voices:
			self.add_voice (Voice (event.voice))
			
		self.voices[event.voice].add_event (event)

	def time_at_measure (self, number):
		if number > len (self):
			raise Exception ("OutOfRangeMeasureError")

		for v in self.voices:
			if self.voices[v].has_measure (number):
				return self.voices[v].measures[number].time

		return None

	def divisions_at_measure (self, number):
		if number > len (self):
			raise Exception ("OutOfRangeMeasureError")

		for v in self.voices:
			if self.voices[v].has_measure (number):
				return self.voices[v].measures[number].divisions

		return None

	def is_pickup (self):
		pad = 0
		
		for i in self.voices:
			if self.voices[i].has_measure (0):
				last_chord = self.voices[i].measures[0].last_chord ()
				pad = last_chord.moment + last_chord.duration
				
				return pad

			# SharpEye's pickup measures are numbered "1"
			try:
				if 'SharpEye' in self.score.identification.encoding['software'][0]:
					if self.voices[i].has_measure (1):
						last_chord = self.voices[i].measures[1].last_chord ()
						pad = last_chord.moment + last_chord.duration
						
						return pad
			except:
				pass
			

		return False

	def __len__ (self):
		m = 0
		for i in self.voices:
			n = max (self.voices[i].measures.keys ())
			if n > m:
				m = n

		return m


class Voice (Object):
	def __init__ (self, number = 1):
		self.number = number
		self.lyrics = None
		self.staves = 1
		self.part = None
		self.measures = {}

		self.__notes_on_staff = {1 : 0, 2 : 0}

	def add_measure (self, measure):
		measure.voice = self
		self.measures[measure.number] = measure

	def add_event (self, event):
		# previous measure - if any
		if len (self.measures.keys ()) > 0:
			prev = max (self.measures.keys ())
		
		if event.measure.number not in self.measures:
			self.add_measure (Measure (event.measure.number))

		if isinstance (event, BaseNote):
			if not hasattr (event, 'chord'):
				if len (self.measures[event.measure.number].events) > 0:
					event.prev = self.measures[event.measure.number].events[-1]
				elif len (self.measures.keys ()) > 1:
					event.prev = self.measures[prev].events[-1]
					self.measures[prev].events[-1].next = event

				event = Chord (event)

		if len (self.measures[event.measure.number].events) > 0:
			event.prev = self.measures[event.measure.number].events[-1]
			self.measures[event.measure.number].events[-1].next = event
		elif len (self.measures.keys ()) > 1:
			event.prev = self.measures[prev].events[-1]
			self.measures[prev].events[-1].next = event


		if event.staff == 0:
			pass
		elif event.staff == 1:
			self.__notes_on_staff[1] += 1
		elif event.staff == 2:
			self.__notes_on_staff[2] += 1
			self.part.staves = 2
		else:
			raise Exception ("UnsupportedNumberOfStavesError")
		
		self.measures[event.measure.number].add_event (event)

	def has_measure (self, number):
		if number in self.measures:
			return True

		return False

	def staff (self):
		staff = 1
		notes = 0
		for n in self.__notes_on_staff:
			if self.__notes_on_staff[n] > notes:
				notes = self.__notes_on_staff[n]
				staff = n
			
		return staff


class Measure (Object):
	def __init__ (self, number = -1):
		self.number = number
		self.voice = None
		self.time = None
		self.divisions = []
		self.events = []

	def add_event (self, event):
		event.measure = self
		
		if len (self.events) > 0 and hasattr (event, 'chord'):
			if isinstance (self.events[-1], Chord):
				self.events[-1].add_note (event)
				return
			else:
				raise Exception ('OrphanChordNoteError')

		if not self.first_chord ():
			self.time = event.time

		if event.divisions and event.divisions not in self.divisions:
			self.divisions.append (event.divisions)
			if len (self.divisions) > 1:
				raise Exception ("FIXME")

		self.events.append (event)

	def first_chord (self):
		for e in self.events:
			if isinstance (e, Chord):
				return e

		return None

	def last_chord (self):
		events = copy.copy (self.events)
		events.reverse ()
		for e in events:
			if isinstance (e, Chord):
				return e
		
		return None

	def events_at (self, start, end = -1):
		events = []

		if end == -1:
			for e in self.events:
				if start == e.moment:
					events.append (e)
		else:
			for e in self.events:
				if start <= e.moment < end:
					events.append (e)

		return events

	def divisions_at (self, start, end = -1):
		divisions = []

		def is_division (d):
			return isinstance (d, Divisions)
		
		events = filter (is_division, self.events_at (start, end))

		for d in self.divisions:
			if d not in events:
				divisions.append (d)
		divisions += events

		return divisions

	def directions_at (self, start, end = -1):
		directions = []

		def is_direction (d):
			return isinstance (d, Direction)
		
		directions = filter (is_direction, self.events_at (start, end))

		return directions

	def timeslices (self, start, stop):
		timeslices = []
		divisions = self.divisions_at (start, stop)
	
		# check when "divisions" attributes happen, but only when
		# there are at least two <divisions>
		# BUG: should work only if there is one <divisions> change
		# in the time slice
		# BUG: fix the first instance of a voice!
		# BUGGY!!!
		if len (divisions) == 0:
			raise Exception ("NoDivisionsError")
		elif len (divisions) == 1:
			dur = stop - start
			div = divisions[0].number
			timeslices.append ({'duration' : dur, 'divisions' : div})			
		elif len (divisions) == 2:
			for d in range (len (divisions)):
				if d == 0:
					dur = divisions[d + 1].moment - start
					div = divisions[d].number
				else:
					dur = stop - divisions[d].moment
					div = divisions[d].number
				timeslices.append ({'duration' : dur, 'divisions' : div})
		else:
			raise Exception ("TooManyDivisionsChangesError")
	
		return timeslices

	def prev (self):
		if (self.number - 1) in self.voice.measures:
			return self.voice.measures[self.number - 1]

		return None
	

	def next (self):
		if (self.number + 1) in self.voice.measures:
			return self.voice.measures[self.number + 1]

		return None

	def __len__ (self):
		if self.time is None:
			return self.divisions[0].number * 4
		
		return self.divisions[0].number * self.time.beats[0] * 4 / self.time.beat_type[0]


# music events: chord (note), direction, attributes, barline (measure)
class Event (Object):
	def __init__ (self, treefrag = None, moment = -1):
		self.moment = moment
		self.duration = 0
		
		self.voice = 1
		self.measure = -1
		self.staff = 1
		
		self.prev = None
		self.next = None

		self.divisions = 0
		self.time = None
		self.clef = None

	def next_note (self):
		next = self.next
		while next is not None:
			if isinstance (next, Chord):
				return next.notes[0]
			next = next.next
			
		return None

	def prev_note (self):
		prev = self.prev
		while prev is not None:
			if isinstance (prev, Chord):
				return prev.notes[0]
			prev = prev.prev

		return None

	def prev_pitch (self):
		prev = self.prev
		while prev is not None:
			if isinstance (prev, Chord) and hasattr (prev.notes[0], 'pitch'):
				return prev.notes[0].pitch
			prev = prev.prev

		return None

	def next_chord (self):
		next = self.next
		while next is not None:
			if isinstance (next, Chord):
				return next
			next = next.next
			
		return None

	def prev_chord (self):
		prev = self.prev
		while prev is not None:
			if isinstance (prev, Chord):
				return prev
			prev = prev.prev

		return None

	def next_chord_not_grace_not_cue (self):
		next = self.next
		while next is not None:
			if isinstance (next, Chord) and isinstance (next.notes[0], Note):
				return next
			next = next.next
			
		return None

	def prev_chord_not_grace_not_cue (self):
		prev = self.prev
		while prev is not None:
			if isinstance (prev, Chord) and isinstance (prev.notes[0], Note):
				return prev
			prev = prev.prev

		return None


	def prev_clef (self):
		prev = self.prev
		while prev is not None:
			if isinstance (prev, Clef):
				return prev
			prev = prev.prev

		return None
		

class Chord (Event):
	def __init__ (self, note = None):
		Event.__init__ (self)

		self.notes = []
		
		if note:
			self.notes.append (note)
			self.moment = note.moment
			self.duration = note.duration
			self.measure = note.measure
			self.time = note.time
			self.divisions = note.divisions
			self.staff = note.staff

	def add_note (self, note):
		if len (self.notes) > 0:
			assert self.duration >= note.duration
		
		self.notes.append (note)
		self.next = note.next
		
		if len (self.notes) == 1:
			self.moment = note.moment
			self.duration = note.duration
			self.measure = note.measure
			self.time = note.time
			self.divisions = note.divisions
			self.staff = note.staff

	def is_first_chord_in_measure (self):
		prev = self.prev_chord ()

		if not prev or prev.measure < self.measure:
			return True

		return False

	def is_grace (self):
		return isinstance (self.notes[0], Grace)


class BaseNote (Event):
	def __init__ (self, treefrag, moment = -1):
		Event.__init__ (self, treefrag, moment)

		attrs = [['pitch', Pitch],                         \
			 ['rest'],                                 \
			 ['unpitched'],                            \
			 ['chord'],                                \
			 ['instrument'],                           \
			 ['voice', 1],                             \
			 ['type', ''],                             \
			 ['dot', len, 0],                          \
			 ['accidental', ''],                       \
			 ['time-modification', TimeModification],  \
			 ['stem', None],                           \
			 ['notehead'],                             \
			 ['staff', 1]]

		set_attributes (self, attrs, treefrag)

		self.time = None
		self.divisions = 0

		self.footnote = ''
		self.level = ''
		self.beam = 0

		self.notations = []
		self.ties = []
		self.slurs = []
		self.articulations = []
		self.ornaments = []
		self.technicals = []
		try:
			for t in treefrag['notations'][0]:
				for i in treefrag['notations'][0][t]:
					if t == 'tied':
						notation = Tied (i)
						self.ties.append (notation)
					elif t == 'slur':
						notation = Slur (i)
						self.slurs.append (notation)
					elif t == 'articulations':
						notation = Articulations (i)
						self.articulations.append (notation)
					elif t == 'ornaments':
						notation = Ornaments (i)
						self.ornaments.append (notation)
					elif t == 'technical':
						notation = Technical (i)
						self.technicals.append (notation)
					else:
						continue
					
					self.notations.append (notation)
		except KeyError:
			pass


class Grace (BaseNote):
	def __init__ (self, treefrag, moment = -1):
		BaseNote.__init__ (self, treefrag, moment)

		attrs = [['tie', 0],        \
			 ['slash', None]]
		
		set_attributes (self, attrs, treefrag)


class Cue (BaseNote):
	def __init__ (self, treefrag, moment = -1):
		BaseNote.__init__ (self, treefrag, moment)

		self.duration = treefrag['duration'][0]


class Note (BaseNote):
	def __init__ (self, treefrag, moment = -1):
		BaseNote.__init__ (self, treefrag, moment)

		self.duration = treefrag['duration'][0]
		
		try:
			self.tie = len (treefrag['tie'])
		except KeyError:
			self.tie = 0


class Pitch (Object):
	def __init__ (self, treefrag = None):
		attrs = [['step'],      \
			 ['alter', 0],  \
			 ['octave']]

		if treefrag:
			set_attributes (self, attrs, treefrag[0])

	def __sub__ (self, pitch):
		steps = ['C', 'D', 'E', 'F', 'G', 'A', 'B']
		a = ((int (self.octave) + 1) * 7) + steps.index (self.step)
		b = ((int (pitch.octave) + 1) * 7) + steps.index (pitch.step)
		if a - b == 0:
			return 0
		return - (abs (b - a) + 1) * (b - a) / abs (b - a)


class TimeModification (Object):
	def __init__ (self, treefrag):
		attrs = [['actual-notes'],    \
			 ['normal-notes'],    \
			 ['normal-type', ''], \
			 ['normal-dot', 0]]

		set_attributes (self, attrs, treefrag[0])

		self.type = None


class Notations (Object):
	def __init__ (self, placement = ''):
		self.placement = placement


class Tied (Notations):
	def __init__ (self, treefrag):
		Notations.__init__ (self)
		
		self.type = treefrag['type']

		try:
			self.number = treefrag['number']
		except KeyError:
			self. number = 1


class Slur (Notations):
	def __init__ (self, treefrag):
		Notations.__init__ (self)

		self.type = treefrag['type']

		try:
			self.number = int (treefrag['number'])
		except KeyError:
			self.number = 1
	

class Articulations (Notations):
	def __init__ (self, treefrag):
		Notations.__init__ (self)

		self.values = {}

		for t in treefrag:
			if t == 'other-articulation':
				if t in self.values:
					self.values[t].append (treefrag[t])
				else:
					self.values[t] = []
					self.values[t].append (treefrag[t])
			else:
				self.values[t] = len (treefrag[t])

	
class Ornaments (Notations):
	def __init__ (self, treefrag):
		Notations.__init__ (self)

		self.values = {}

		for t in treefrag:
			if t == 'other-ornament':
				if t in self.values:
					self.values[t].append (treefrag[t])
				else:
					self.values[t] = []
					self.values[t].append (treefrag[t])
			else:
				self.values[t] = len (treefrag[t])


class Technical (Notations):
	def __init__ (self, treefrag):
		Notations.__init__ (self)

		self.values = {}

		for t in treefrag:
			if t == 'other-technical':
				if t in self.values:
					self.values[t].append (treefrag[t])
				else:
					self.values[t] = []
					self.values[t].append (treefrag[t])
			else:
				self.values[t] = len (treefrag[t])


class Lyric (Object):
	def __init__ (self, number = 1, name = ''):
		self.number = number
		self.name = ''
		self.text = ''
		self.syllabic = ''
		self.elision = False
		self.extend = False

	def __repr__ (self):
		return '<' + self.text + '>'

	def set_extend (self):
		if self.syllabic in ['begin', 'middle']:
			raise Exception ('ExtendOnStartingSyllabeError')
		self.extend = True

	def set_text (self, text):
		#sys.stderr.write ('[' + text + ']')
		self.text = text

	def set_syllabic (self, syllabic):
		if syllabic not in ['single', 'begin', 'end', 'middle']:
			raise Exception ('UnknownSyllabicError')
		self.syllabic = syllabic


class Divisions (Event):
	def __init__ (self, treefrag, moment = 0):
		Event.__init__ (self, treefrag, moment)
		
		self.number = treefrag


class Time (Event):
	def __init__ (self, treefrag, moment = 0):
		Event.__init__ (self, treefrag, moment)

		attrs = [['number', 0]]
		
		set_attributes (self, attrs, treefrag)

		self.staff = self.number

		if treefrag.has_key ('beats') and treefrag.has_key ('beat-type'):
			self.beats = treefrag['beats']
			self.beat_type = treefrag['beat-type']
		elif treefrag.has_key ('senza-misura'):
			self.senza_misura = True


class Key (Event):
	def __init__ (self, treefrag, moment = 0):
		Event.__init__ (self, treefrag, moment)
		
		attrs = [['fifths', 0],       \
			 ['mode', 'major'],   \
			 ['number', 0]]

		set_attributes (self, attrs, treefrag)

		self.staff = self.number


class Clef (Event):
	def __init__ (self, treefrag, moment = 0):
		Event.__init__ (self, treefrag, moment)

		attrs = [['sign'],              \
			 ['line', 0],           \
			 ['octave-change', 0],  \
			 ['number', 1]]

		set_attributes (self, attrs, treefrag)

		self.staff = self.number


class Direction (Event):
	def __init__ (self, treefrag, moment = -1):
		Event.__init__ (self, treefrag, moment)

		attrs = [['voice', 1],      \
			 ['staff', 1],      \
			 ['offset', 0]]

		set_attributes (self, attrs, treefrag)


class Rehearsal (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Segno (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Words (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Coda (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Wedge (Direction):
	def __init__ (self, treefrag, moment = -1):
		Direction.__init__ (self, treefrag, moment)

		attrs = [['type', str],   \
			 ['number', 1],   \
			 ['spread', 0]]

		set_attributes (self, attrs, treefrag)


class Dynamics (Direction):
	def __init__ (self, treefrag, moment = -1):
		Direction.__init__ (self, treefrag, moment)

		self.values = []

		for i in treefrag:
			if i in ['p', 'pp', 'ppp', 'pppp', 'ppppp', 'pppppp', \
				 'f', 'ff', 'fff', 'ffff', 'fffff', 'ffffff', \
				 'mp', 'mf', 'sf', 'sfp', 'sfpp', 'fp',       \
				 'rf', 'rfz', 'sfz', 'sffz', 'fz']:
				self.values.append (i)
			elif i == 'other-dynamics':
				self.values.append (treefrag['other-dynamics'])


class Dashes (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Bracket (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Pedal (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__

		self.type = ''
		self.line = 'yes'


class Metronome (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class OctaveShift (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__

		self.type = ''
		self.number = 1
		self.size = 8


class HarpPedals (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Damp (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class DampAll (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Eyeglasses (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class Scordatura (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__


class OtherDirection (Direction):
	def __init__ (self, direction = None):
		Direction.__init__ (self)
		# kludge?
		if direction is not None:
			self.__dict__ = direction.__dict__

class Technical (Notations):
	def set_type (type, substitution = 'no'):
		self.type = type
		self.substitution = substitution

	def set_fingering (finger):
		if self.type != 'fingering':
			raise Exception ('FingerOnNotFingeringError')
		self.finger = finger


class Barline (Event):
	def __init__ (self, treefrag, moment = -1):
		Event.__init__ (self, treefrag)




def NoteFactory (treefrag, moment):
	if treefrag.has_key ('grace'):
		return Grace (treefrag, moment)
	elif treefrag.has_key ('cue'):
		return Cue (treefrag, moment)
	else:
		return Note (treefrag, moment)


def AttributesFactory (treefrag, moment):
	attributes = []
	
	for key in treefrag:
		if key == 'divisions':
			attributes.append (Divisions (treefrag[key][0], moment))
		elif key == 'key':
			attributes.append (Key (treefrag[key][0], moment))
		elif key == 'time':
			attributes.append (Time (treefrag[key][0], moment))
		elif key == 'clef':
# 			for f in treefrag[key]:
# 				attributes.append (Clef (f, moment))
				attributes.append (Clef (treefrag[key][0], moment))

	return attributes


def DirectionFactory (treefrag, moment):
	directions = []

	for dirtype in treefrag['direction-type']:
		for direction in dirtype:
			if direction == 'dynamics':
				directions.append (Dynamics (dirtype[direction][0], moment))
			elif direction == 'wedge':
				directions.append (Wedge (dirtype[direction][0], moment))

	return directions


def BarlineFactory (treefrag, moment = 0):
	return Barline (treefrag, moment)

